------------------------------------------------------------------------------
-- LED (80bit key, enc only)
------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity led80 is
	port (
		din		: in  std_logic_vector(63 downto 0);
		dout	: out std_logic_vector(63 downto 0);
		kin		: in  std_logic_vector(79 downto 0);

		start	: in  std_logic;
		busy	: out std_logic;

		rst		: in  std_logic;	-- sync reset, active H
		clk		: in  std_logic

	);
end led80;

architecture RTL of led80 is
	------------------------------------------------------------/
	-- internal wires/registers
	------------------------------------------------------------/
	signal data_reg		: std_logic_vector(63 downto 0);
	signal key_reg		: std_logic_vector(79 downto 0);
	signal rnd_reg		: std_logic_vector(5 downto 0);

	constant ST_WAIT	: std_logic := '0';
	constant ST_CALC0	: std_logic := '1';
	signal state_reg	: std_logic;

	------------------------------------------------------------/
	-- functions
	------------------------------------------------------------/
	function AddConstants_func (
		p	: std_logic_vector(63 downto 0);
		rnd	: std_logic_vector(5 downto 0)
	) return std_logic_vector is
		variable const					: std_logic_vector(5 downto 0);
		variable c0, c1, c2, c3			: std_logic_vector(3 downto 0);
		variable AddConstants_func_ret	: std_logic_vector(63 downto 0);
	begin
		case (rnd) is
		when CONV_STD_LOGIC_VECTOR(0,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"01")),6);
		when CONV_STD_LOGIC_VECTOR(1,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"03")),6);
		when CONV_STD_LOGIC_VECTOR(2,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"07")),6);
		when CONV_STD_LOGIC_VECTOR(3,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"0F")),6);
		when CONV_STD_LOGIC_VECTOR(4,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"1F")),6);
		when CONV_STD_LOGIC_VECTOR(5,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"3E")),6);
		when CONV_STD_LOGIC_VECTOR(6,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"3D")),6);
		when CONV_STD_LOGIC_VECTOR(7,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"3B")),6);
		when CONV_STD_LOGIC_VECTOR(8,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"37")),6);
		when CONV_STD_LOGIC_VECTOR(9,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"2F")),6);
		when CONV_STD_LOGIC_VECTOR(10,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"1E")),6);
		when CONV_STD_LOGIC_VECTOR(11,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"3C")),6);
		when CONV_STD_LOGIC_VECTOR(12,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"39")),6);
		when CONV_STD_LOGIC_VECTOR(13,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"33")),6);
		when CONV_STD_LOGIC_VECTOR(14,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"27")),6);
		when CONV_STD_LOGIC_VECTOR(15,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"0E")),6);
		when CONV_STD_LOGIC_VECTOR(16,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"1D")),6);
		when CONV_STD_LOGIC_VECTOR(17,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"3A")),6);
		when CONV_STD_LOGIC_VECTOR(18,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"35")),6);
		when CONV_STD_LOGIC_VECTOR(19,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"2B")),6);
		when CONV_STD_LOGIC_VECTOR(20,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"16")),6);
		when CONV_STD_LOGIC_VECTOR(21,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"2C")),6);
		when CONV_STD_LOGIC_VECTOR(22,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"18")),6);
		when CONV_STD_LOGIC_VECTOR(23,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"30")),6);
		when CONV_STD_LOGIC_VECTOR(24,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"21")),6);
		when CONV_STD_LOGIC_VECTOR(25,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"02")),6);
		when CONV_STD_LOGIC_VECTOR(26,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"05")),6);
		when CONV_STD_LOGIC_VECTOR(27,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"0B")),6);
		when CONV_STD_LOGIC_VECTOR(28,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"17")),6);
		when CONV_STD_LOGIC_VECTOR(29,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"2E")),6);
		when CONV_STD_LOGIC_VECTOR(30,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"1C")),6);
		when CONV_STD_LOGIC_VECTOR(31,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"38")),6);
		when CONV_STD_LOGIC_VECTOR(32,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"31")),6);
		when CONV_STD_LOGIC_VECTOR(33,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"23")),6);
		when CONV_STD_LOGIC_VECTOR(34,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"06")),6);
		when CONV_STD_LOGIC_VECTOR(35,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"0D")),6);
		when CONV_STD_LOGIC_VECTOR(36,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"1B")),6);
		when CONV_STD_LOGIC_VECTOR(37,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"36")),6);
		when CONV_STD_LOGIC_VECTOR(38,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"2D")),6);
		when CONV_STD_LOGIC_VECTOR(39,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"1A")),6);
		when CONV_STD_LOGIC_VECTOR(40,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"34")),6);
		when CONV_STD_LOGIC_VECTOR(41,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"29")),6);
		when CONV_STD_LOGIC_VECTOR(42,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"12")),6);
		when CONV_STD_LOGIC_VECTOR(43,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"24")),6);
		when CONV_STD_LOGIC_VECTOR(44,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"08")),6);
		when CONV_STD_LOGIC_VECTOR(45,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"11")),6);
		when CONV_STD_LOGIC_VECTOR(46,6) =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"22")),6);
		when others =>
			const	:= CONV_STD_LOGIC_VECTOR(CONV_INTEGER(To_StdLogicVector(X"04")),6);
		end case;

		c0	:= ('0' & const(5) & const(4) & const(3));
		c1	:= ('0' & const(2) & const(1) & const(0));
		c2	:= ('0' & const(5) & const(4) & const(3));
		c3	:= ('0' & const(2) & const(1) & const(0));

		AddConstants_func_ret	:= ((p(63 downto 60) xor CONV_STD_LOGIC_VECTOR(5,4)) 
									& (p(59 downto 56) xor c0) & p(55 downto 52) & p(51 downto 48) 
									& (p(47 downto 44) xor CONV_STD_LOGIC_VECTOR(1,4) xor CONV_STD_LOGIC_VECTOR(5,4)) 
									& (p(43 downto 40) xor c1) & p(39 downto 36) & p(35 downto 32) 
									& (p(31 downto 28) xor CONV_STD_LOGIC_VECTOR(2,4)) & (p(27 downto 24) xor c2) 
									& p(23 downto 20) & p(19 downto 16) & (p(15 downto 12) xor CONV_STD_LOGIC_VECTOR(3,4)) 
									& (p(11 downto 8) xor c3) & p(7 downto 4) & p(3 downto 0));
		return AddConstants_func_ret;
	end AddConstants_func;

	-- sbox
	function s_func (
		i	: std_logic_vector(3 downto 0)
	) return std_logic_vector is
		variable o			: std_logic_vector(3 downto 0);
		variable s_func_ret	: std_logic_vector(3 downto 0);
	begin
		case (i) is
		when To_StdLogicVector(X"0") =>
			o	:= To_StdLogicVector(X"C");
		when To_StdLogicVector(X"1") =>
			o	:= To_StdLogicVector(X"5");
		when To_StdLogicVector(X"2") =>
			o	:= To_StdLogicVector(X"6");
		when To_StdLogicVector(X"3") =>
			o	:= To_StdLogicVector(X"B");
		when To_StdLogicVector(X"4") =>
			o	:= To_StdLogicVector(X"9");
		when To_StdLogicVector(X"5") =>
			o	:= To_StdLogicVector(X"0");
		when To_StdLogicVector(X"6") =>
			o	:= To_StdLogicVector(X"A");
		when To_StdLogicVector(X"7") =>
			o	:= To_StdLogicVector(X"D");
		when To_StdLogicVector(X"8") =>
			o	:= To_StdLogicVector(X"3");
		when To_StdLogicVector(X"9") =>
			o	:= To_StdLogicVector(X"E");
		when To_StdLogicVector(X"A") =>
			o	:= To_StdLogicVector(X"F");
		when To_StdLogicVector(X"B") =>
			o	:= To_StdLogicVector(X"8");
		when To_StdLogicVector(X"C") =>
			o	:= To_StdLogicVector(X"4");
		when To_StdLogicVector(X"D") =>
			o	:= To_StdLogicVector(X"7");
		when To_StdLogicVector(X"E") =>
			o	:= To_StdLogicVector(X"1");
		when others =>
			o	:= To_StdLogicVector(X"2");
		end case;

		s_func_ret	:= o;
		return s_func_ret;
	end s_func;

	function SubCells_func (
		p	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable o					: std_logic_vector(63 downto 0);
		variable SubCells_func_ret	: std_logic_vector(63 downto 0);
	begin
		o(63 downto 60)	:= s_func(p(63 downto 60));
		o(59 downto 56)	:= s_func(p(59 downto 56));
		o(55 downto 52)	:= s_func(p(55 downto 52));
		o(51 downto 48)	:= s_func(p(51 downto 48));
		o(47 downto 44)	:= s_func(p(47 downto 44));
		o(43 downto 40)	:= s_func(p(43 downto 40));
		o(39 downto 36)	:= s_func(p(39 downto 36));
		o(35 downto 32)	:= s_func(p(35 downto 32));
		o(31 downto 28)	:= s_func(p(31 downto 28));
		o(27 downto 24)	:= s_func(p(27 downto 24));
		o(23 downto 20)	:= s_func(p(23 downto 20));
		o(19 downto 16)	:= s_func(p(19 downto 16));
		o(15 downto 12)	:= s_func(p(15 downto 12));
		o(11 downto 8)	:= s_func(p(11 downto 8));
		o(7 downto 4)	:= s_func(p(7 downto 4));
		o(3 downto 0)	:= s_func(p(3 downto 0));

		SubCells_func_ret	:= o;
		return SubCells_func_ret;
	end SubCells_func;

	function ShiftRows_func (
		p	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable ShiftRows_func_ret	: std_logic_vector(63 downto 0);
	begin
		ShiftRows_func_ret	:= (p(63 downto 60) & p(59 downto 56) & p(55 downto 52) & p(51 downto 48) 
								& p(43 downto 40) & p(39 downto 36) & p(35 downto 32) & p(47 downto 44) 
								& p(23 downto 20) & p(19 downto 16) & p(31 downto 28) & p(27 downto 24) 
								& p(3 downto 0) & p(15 downto 12) & p(11 downto 8) & p(7 downto 4));
		return ShiftRows_func_ret;
	end ShiftRows_func;

	-- mult on GF(2^4), x^4+x+1
	function mult_func (
		a	: std_logic_vector(3 downto 0);
		b	: std_logic_vector(3 downto 0)
	) return std_logic_vector is
		variable o				: std_logic_vector(6 downto 0);
		variable p				: std_logic_vector(3 downto 0);
		variable mult_func_ret	: std_logic_vector(3 downto 0);
	begin
		o(0)	:= (a(0) and b(0));
		o(1)	:= (a(1) and b(0)) xor (a(0) and b(1));
		o(2)	:= (a(2) and b(0)) xor (a(1) and b(1)) xor (a(0) and b(2));
		o(3)	:= (a(3) and b(0)) xor (a(2) and b(1)) xor (a(1) and b(2)) xor (a(0) and b(3));
		o(4)	:= (a(3) and b(1)) xor (a(2) and b(2)) xor (a(1) and b(3));
		o(5)	:= (a(3) and b(2)) xor (a(2) and b(3));
		o(6)	:= (a(3) and b(3));

		p(0)	:= o(0) xor o(4);
		p(1)	:= o(1) xor o(4) xor o(5);
		p(2)	:= o(2) xor o(5) xor o(6);
		p(3)	:= o(3) xor o(6);

		mult_func_ret	:= p;
		return mult_func_ret;
	end mult_func;

	-- row-column multiplication
	function rcm_func (
		a	: std_logic_vector(15 downto 0);
		b	: std_logic_vector(15 downto 0)
	) return std_logic_vector is
		variable a0, a1, a2, a3	: std_logic_vector(3 downto 0);
		variable b0, b1, b2, b3	: std_logic_vector(3 downto 0);
		variable rcm_func_ret	: std_logic_vector(3 downto 0);
	begin
		a3	:= a(15 downto 12);
		a2	:= a(11 downto 8);
		a1	:= a(7 downto 4);
		a0	:= a(3 downto 0);

		b3	:= b(15 downto 12);
		b2	:= b(11 downto 8);
		b1	:= b(7 downto 4);
		b0	:= b(3 downto 0);

		rcm_func_ret	:= mult_func(a3, b3) xor mult_func(a2, b2) xor mult_func(a1, b1) xor mult_func(a0, b0);
		return rcm_func_ret;
	end rcm_func;

	function MixColumnsSerial_func (
		p	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable r0, r1, r2, r3				: std_logic_vector(15 downto 0);
		variable c0, c1, c2, c3				: std_logic_vector(15 downto 0);
		variable MixColumnsSerial_func_ret	: std_logic_vector(63 downto 0);
	begin
		r0	:= (p(63 downto 60) & p(47 downto 44) & p(31 downto 28) & p(15 downto 12));
		r1	:= (p(59 downto 56) & p(43 downto 40) & p(27 downto 24) & p(11 downto 8));
		r2	:= (p(55 downto 52) & p(39 downto 36) & p(23 downto 20) & p(7 downto 4));
		r3	:= (p(51 downto 48) & p(35 downto 32) & p(19 downto 16) & p(3 downto 0));

		c0	:= (To_StdLogicVector(X"4") & To_StdLogicVector(X"1") & To_StdLogicVector(X"2") & To_StdLogicVector(X"2"));
		c1	:= (To_StdLogicVector(X"8") & To_StdLogicVector(X"6") & To_StdLogicVector(X"5") & To_StdLogicVector(X"6"));
		c2	:= (To_StdLogicVector(X"B") & To_StdLogicVector(X"E") & To_StdLogicVector(X"A") & To_StdLogicVector(X"9"));
		c3	:= (To_StdLogicVector(X"2") & To_StdLogicVector(X"2") & To_StdLogicVector(X"F") & To_StdLogicVector(X"B"));

		MixColumnsSerial_func_ret	:= (rcm_func(c0, r0) & rcm_func(c0, r1) & rcm_func(c0, r2) & rcm_func(c0, r3) 
										& rcm_func(c1, r0) & rcm_func(c1, r1) & rcm_func(c1, r2) & rcm_func(c1, r3) 
										& rcm_func(c2, r0) & rcm_func(c2, r1) & rcm_func(c2, r2) & rcm_func(c2, r3) 
										& rcm_func(c3, r0) & rcm_func(c3, r1) & rcm_func(c3, r2) & rcm_func(c3, r3));
		return MixColumnsSerial_func_ret;
	end MixColumnsSerial_func;

	signal dk, p0, p1, p2, p3, next_data_node	: std_logic_vector(63 downto 0);
	signal kn	: std_logic_vector(63 downto 0);


begin

	kn	<= key_reg(79 downto 16)	when (rnd_reg(2) = '0')	else (key_reg(15 downto 0) & CONV_STD_LOGIC_VECTOR(0,48));
	dk	<= data_reg xor kn;

	p0	<= dk	when (rnd_reg(1 downto 0) = CONV_STD_LOGIC_VECTOR(0,2))	else data_reg;
	p1	<= AddConstants_func(p0, rnd_reg);
	p2	<= SubCells_func(p1);
	p3	<= ShiftRows_func(p2);

	next_data_node	<= MixColumnsSerial_func(p3);

	------------------------------------------------------------/
	-- Main control
	------------------------------------------------------------/
	pr0:process (clk)
	begin
		if (clk'event and clk = '1') then
			if (rst = '1') then
				state_reg	<= ST_WAIT;

			else
				case (state_reg) is
				--------------------------------
				-- wait trigger
				--------------------------------
				when ST_WAIT =>
					if (start = '1') then
						key_reg		<= kin;
						data_reg	<= din;
						rnd_reg		<= CONV_STD_LOGIC_VECTOR(0,6);
						state_reg	<= ST_CALC0;
					end if;

				--------------------------------
				-- ENC
				--------------------------------
				when ST_CALC0 =>
					rnd_reg	<= rnd_reg + '1';

					if (rnd_reg <= CONV_STD_LOGIC_VECTOR(47,6)) then
						data_reg	<= next_data_node;
					else
						data_reg	<= dk;						-- final addRoundKey
						state_reg	<= ST_WAIT;
					end if;

				when others =>
					null;
				end case;
			end if;
		end if;
	end process;

	------------------------------------------------------------/
	-- PO
	------------------------------------------------------------/
	dout	<= data_reg;
	busy	<= state_reg;

end RTL;
