------------------------------------------------------------------------------
-- Present (80bit key, enc only)
------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity present80 is
	port (
		din		: in  std_logic_vector(63 downto 0);
		dout	: out std_logic_vector(63 downto 0);
		kin		: in  std_logic_vector(79 downto 0);

		start	: in  std_logic;
		busy	: out std_logic;

		rst		: in  std_logic;	-- sync reset, active H
		clk		: in  std_logic
	);
end present80;

architecture RTL of present80 is
	------------------------------------------------------------/
	-- internal wires/registers
	------------------------------------------------------------/
	signal data_reg			: std_logic_vector(63 downto 0);
	signal round_key_reg	: std_logic_vector(79 downto 0);
	signal rc_reg			: std_logic_vector(5 downto 0);

	signal next_data		: std_logic_vector(63 downto 0);
	signal next_key			: std_logic_vector(79 downto 0);
	signal ek_node			: std_logic_vector(63 downto 0);
	signal next_rc			: std_logic_vector(5 downto 0);

	constant ST_WAIT		: std_logic := '0';
	constant ST_CALC0		: std_logic := '1';
	signal state_reg		: std_logic;

	------------------------------------------------------------/
	-- functions
	------------------------------------------------------------/

	----------------
	-- S
	----------------
	function s_func (
		i	: std_logic_vector(3 downto 0)
	) return std_logic_vector is
		variable o			: std_logic_vector(3 downto 0);
		variable s_func_ret	: std_logic_vector(3 downto 0);
	begin
		case (i) is
		when To_StdLogicVector(X"0") =>
			o	:= To_StdLogicVector(X"C");
		when To_StdLogicVector(X"1") =>
			o	:= To_StdLogicVector(X"5");
		when To_StdLogicVector(X"2") =>
			o	:= To_StdLogicVector(X"6");
		when To_StdLogicVector(X"3") =>
			o	:= To_StdLogicVector(X"B");
		when To_StdLogicVector(X"4") =>
			o	:= To_StdLogicVector(X"9");
		when To_StdLogicVector(X"5") =>
			o	:= To_StdLogicVector(X"0");
		when To_StdLogicVector(X"6") =>
			o	:= To_StdLogicVector(X"A");
		when To_StdLogicVector(X"7") =>
			o	:= To_StdLogicVector(X"D");
		when To_StdLogicVector(X"8") =>
			o	:= To_StdLogicVector(X"3");
		when To_StdLogicVector(X"9") =>
			o	:= To_StdLogicVector(X"E");
		when To_StdLogicVector(X"A") =>
			o	:= To_StdLogicVector(X"F");
		when To_StdLogicVector(X"B") =>
			o	:= To_StdLogicVector(X"8");
		when To_StdLogicVector(X"C") =>
			o	:= To_StdLogicVector(X"4");
		when To_StdLogicVector(X"D") =>
			o	:= To_StdLogicVector(X"7");
		when To_StdLogicVector(X"E") =>
			o	:= To_StdLogicVector(X"1");
		when others =>
			o	:= To_StdLogicVector(X"2");
		end case;

		s_func_ret	:= o;
		return s_func_ret;
	end s_func;

	----------------
	-- S-array
	----------------
	function s_ary_func (
		i	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable o				: std_logic_vector(63 downto 0);
		variable s_ary_func_ret	: std_logic_vector(63 downto 0);
	begin
		o(63 downto 60)	:= s_func(i(63 downto 60));
		o(59 downto 56)	:= s_func(i(59 downto 56));
		o(55 downto 52)	:= s_func(i(55 downto 52));
		o(51 downto 48)	:= s_func(i(51 downto 48));
		o(47 downto 44)	:= s_func(i(47 downto 44));
		o(43 downto 40)	:= s_func(i(43 downto 40));
		o(39 downto 36)	:= s_func(i(39 downto 36));
		o(35 downto 32)	:= s_func(i(35 downto 32));
		o(31 downto 28)	:= s_func(i(31 downto 28));
		o(27 downto 24)	:= s_func(i(27 downto 24));
		o(23 downto 20)	:= s_func(i(23 downto 20));
		o(19 downto 16)	:= s_func(i(19 downto 16));
		o(15 downto 12)	:= s_func(i(15 downto 12));
		o(11 downto 8)	:= s_func(i(11 downto 8));
		o(7 downto 4)	:= s_func(i(7 downto 4));
		o(3 downto 0)	:= s_func(i(3 downto 0));
		s_ary_func_ret	:= o;
		return s_ary_func_ret;
	end s_ary_func;

	----------------
	-- P
	----------------
	function p_func (
		i	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable o			: std_logic_vector(63 downto 0);
		variable p_func_ret	: std_logic_vector(63 downto 0);
	begin
		o	:= (i(63) & i(59) & i(55) & i(51) & i(47) & i(43) & i(39) & i(35) 
				& i(31) & i(27) & i(23) & i(19) & i(15) & i(11) & i(7) & i(3) 
				& i(62) & i(58) & i(54) & i(50) & i(46) & i(42) & i(38) & i(34) 
				& i(30) & i(26) & i(22) & i(18) & i(14) & i(10) & i(6) & i(2) 
				& i(61) & i(57) & i(53) & i(49) & i(45) & i(41) & i(37) & i(33) 
				& i(29) & i(25) & i(21) & i(17) & i(13) & i(9) & i(5) & i(1) 
				& i(60) & i(56) & i(52) & i(48) & i(44) & i(40) & i(36) & i(32) 
				& i(28) & i(24) & i(20) & i(16) & i(12) & i(8) & i(4) & i(0));
		p_func_ret	:= o;
		return p_func_ret;
	end p_func;

	----------------
	-- Round
	----------------
	function round_func (
		p	: std_logic_vector(63 downto 0);
		ek	: std_logic_vector(63 downto 0)
	) return std_logic_vector is
		variable i0				: std_logic_vector(63 downto 0);
		variable round_func_ret	: std_logic_vector(63 downto 0);
	begin
		i0				:= p xor ek;
		round_func_ret	:= p_func(s_ary_func(i0));
		return round_func_ret;
	end round_func;

	----------------
	-- key scheduler
	----------------
	function key_func (
		k	: std_logic_vector(79 downto 0);
		c	: std_logic_vector(5 downto 0)
	) return std_logic_vector is
		variable i0, i1, i2		: std_logic_vector(79 downto 0);
		variable key_func_ret	: std_logic_vector(79 downto 0);
	begin
		i0	:= (k(18 downto 0) & k(79 downto 19));
		i1	:= (s_func(i0(79 downto 76)) & i0(75 downto 0));
		i2	:= (i1(79 downto 20) & (i1(19 downto 15) xor c(4 downto 0)) & i1(14 downto 0) );

		key_func_ret	:= i2;
		return key_func_ret;
	end key_func;

begin
	ek_node		<= round_key_reg(79 downto 16);
	next_data	<= round_func(data_reg, ek_node);
	next_key	<= key_func(round_key_reg, rc_reg);
	next_rc		<= rc_reg + CONV_STD_LOGIC_VECTOR(1,6);

	------------------------------------------------------------/
	-- Main control
	------------------------------------------------------------/
	pr0:process (clk)
	begin
		if (clk'event and clk = '1') then
			if (rst = '1') then
				state_reg	<= ST_WAIT;

			else
				case (state_reg) is
				--------------------------------
				-- wait trigger
				--------------------------------
				when ST_WAIT =>
					if (start = '1') then
						round_key_reg	<= kin;
						data_reg		<= din;
						rc_reg			<= CONV_STD_LOGIC_VECTOR(1,6);
						state_reg		<= ST_CALC0;
					end if;

				--------------------------------
				-- ENC
				--------------------------------
				when ST_CALC0 =>
					if (rc_reg = CONV_STD_LOGIC_VECTOR(32,6)) then
						data_reg		<= data_reg xor ek_node;
						state_reg		<= ST_WAIT;
					else
						data_reg		<= next_data;
						round_key_reg	<= next_key;
						rc_reg			<= next_rc;
					end if;

				when others =>
					null;
				end case;
			end if;
		end if;
	end process;

	------------------------------------------------------------/
	-- PO
	------------------------------------------------------------/
	dout	<= data_reg;
	busy	<= state_reg;

end RTL;
